//---------------------------------------------------------------------------------------
// Визуализация данных на четырех-разрядном 7-сегментном индикаторе
// leddd06 v. 2016.11.30.12.25
// OldBean
//---------------------------------------------------------------------------------------
#define COMMON_CATHODE  // Если индикатор с общим анодом - необходимо закомментировать эту строку
#define DIGS 4          // Количество разрядов в индикаторе (тестировалось только для 3-х или 4-х)
#define DP_POS 2        // Положение десятичной точки. Точка не будет показываться если DP_POS >= DIGS
//---------------------------------------------------------------------------------------
#ifdef COMMON_CATHODE // Для индикаторов с общим катодом
#define L_HIGH 1
#define L_LOW  0
#else                 // Для индикаторов с общим анодом
#define L_HIGH 0
#define L_LOW  1
#endif
//---------------------------------------------------------------------------------------
// Пины, к которым подключены сегменты индикатора аноды (или катоды для индикаторов с общими анодами)
static uint8_t segPins[8] = {12, 16, 5, 7, 8, 13, 4, 6};
// Пины, к которым подключены общие катоды (или аноды, для индикаторов с общими анодами)
static uint8_t digPins[4] = {11, 14, 15, 3}; // Максимальное количество разрядов - 4
static uint8_t digit[12] = {0b00111111, // Маски сегментов цифр 0, 1, 2 и т.д.
                            0b00000110,
                            0b01011011,
                            0b01001111,
                            0b01100110,
                            0b01101101,
                            0b01111101,
                            0b00000111,
                            0b01111111,
                            0b01101111,
                            0b00000000,   // Пробел
                            0b01000000};  // Минус
static int d[DIGS]; // Тысячи, сотни, десятки и единицы, показываемые на разрядах 7-сегментного индикатора
static uint8_t mask = 0x01; // Маска для анализа цифр
static int seg = 0; // Номер сегмента, обрабатываемого (подствечиваемого, если нужно) в данный момент
static unsigned long dt = 3000; // Длительность (в мкс) подсветки каждой группы сегментов
static volatile unsigned long stt = 0; // Момент времени (в мкс) когда заканчивается подсветка текущей группы сегментов
static volatile uint8_t nextSegFlag = 1; // Флаг для перехода к подсветке следующей группы сегментов

static volatile int tval = 0; // Просто число для тестирования
static int maxval; // Максимальное число, отбражаемое на индикаторе + 1
static uint8_t updateFlag = 1; // 1 - данные (цифры), выводимые на индикатор, нужно обновить

// Рабочие переменные для отладки
static volatile unsigned long stg = 0;
static volatile int cntr = 0;
//---------------------------------------------------------------------------------------
void setup() {
  for(int i = 0; i < 8; i++) { // Инициализация пинов, к которым подключены аноды индикатора (сегменты)
    pinMode(segPins[i], OUTPUT);
    digitalWrite(segPins[i], L_LOW);
  }
  for(int i = 0; i < DIGS; i++) { // Инициализация пинов, к которым подключены катоды индикатора (цифры)
    pinMode(digPins[i], OUTPUT);
    digitalWrite(digPins[i], L_HIGH);
    d[i] = 8; // Для демонстрации работоспособности индикатора
  }
  unsigned long st = millis() + 2000;
  while(millis() < st) { // Для демонстрации работоспособности индикатора покажем все сегменты разом
      ShowSegments();
  }
  for(int i = 0; i < DIGS; i++) // Очистим индикатор
    d[i] = 10;
  ShowSegments();

  stg = micros() + dt;

  maxval = 1;
  for(int i = 0; i < DIGS; i++)
    maxval *= 10;
}
//---------------------------------------------------------------------------------------
void loop() {  
  if((++cntr % 2000) == 0) {
    if(++tval >= maxval)
      tval = 0;
    updateFlag = 1; 
  }

  if(updateFlag) // Обновим цифры, выводимые на индикатор
    UpdateDigits(tval);

  if(micros() >= stg) {
    stg = micros() + dt;
    PORTB |= 0b00000010;
    ShowSegments(); // Подсвечиваем очередную группу сегментов. Время отработки функции - 28 мкс.
    PORTB &= 0b11111101;
    
  }
}
//---------------------------------------------------------------------------------------
void UpdateDigits(int val) { // Подготовка данных (цифр) для визуализации
  if(val >= maxval) {
    for(int i = 0; i < DIGS; i++) d[i] = 11; 
    updateFlag = 0;
    return;
  }
      
  int wv = val, dv = maxval/10;
  for(int i = 0; i < DIGS; i++) {
    d[i] = wv/dv; wv -= d[i]*dv; dv /= 10;
  }

  for(int i = 0; i < DP_POS; i++) { // Убираем ведущие нули
    if(d[i] == 0)
      d[i] = 10;
    else
      break;
  }

  updateFlag = 0;
}
//---------------------------------------------------------------------------------------
void ShowSegments() { // Подсвечивание очередной группы сегментов
  digitalWrite(segPins[seg], L_LOW); // "Роняем" потенциал анода "отработавшего" сегмента
  if(++seg < 8) mask <<= 1; // Переходим к следующему сегменту
  else { // Дошли до последнего сегмента
    seg = 0; mask = 0x01;
  }
  digitalWrite(segPins[seg], L_HIGH); // "Подбрасываем" потенциал анода нового текущего сегмента
  for(int bi = 0; bi < DIGS; bi++) { // Теперь пробегаем по цифрам (разрядам) индикатора
    if(seg == 7 && bi == DP_POS) // Подсветим нужную десятичную точку
      digitalWrite(digPins[bi], L_LOW);
    else {
      if(digit[d[bi]] & mask) // Если данный сегмент в bi-ом разряде должен гореть,
        digitalWrite(digPins[bi], L_LOW); // то уроним потенциал катода соответствующей цифры (разряда)
      else // А если гореть не должен, то
        digitalWrite(digPins[bi], L_HIGH); // потенциал катода bi-й цифры должен быть высокий 
    }
  }
}

